import gym
from gym import spaces
from .core import Snake


class SnakeEnv(gym.Env):
    metadata = {"render_modes": ["human"], "render_fps": 30}

    def __init__(self, render_mode=None, **kwargs) -> None:
        self.render_mode = render_mode
        self.action_space = spaces.Discrete(4)
        self.observation_space = spaces.Box(
            float("-inf"), float("inf"), shape=(6,))
        self.snake = Snake(**kwargs)

    def reset(self, seed=None, options=None):
        self.snake.init()
        if self.render_mode == "human":
            self._render_frame()
        return self._get_obs(), self._get_info()

    def step(self, action):
        s, r, d, t = self.snake.step(action)
        if self.render_mode == "human":
            self._render_frame()
        return s, r, d, t, self._get_info()

    def _get_obs(self):
        return self.snake.observation()

    def _get_info(self):
        return self.snake.info()

    def _render_frame(self):
        self.snake.render()

    def close(self):
        self.snake.close()

    def play(self):
        self.snake.play()



from gym_snake_game.version import VERSION as __version__
from gym.envs.registration import register

import gym

__author__ = "Ming Yu"

register(
    id="Snake-v0",
    entry_point="gym_snake_game.environment:SnakeEnv",
)


def make(name, **kwargs):
    return gym.make(name, **kwargs)



import math
import numpy as np
from .utils import *


class Snake:
    def __init__(
        self,
        fps=60,
        max_step=500,
        init_length=4,
        food_reward=2.0,
        dist_reward=None,
        living_bonus=0.0,
        death_penalty=-1.0,
        width=40,
        height=40,
        block_size=20,
        background_color=Color.orange,
        food_color=Color.red,
        head_color=Color.purple,
        body_color=Color.blue,
    ) -> None:

        self.episode = 0
        self.fps = fps
        self.max_step = max_step
        self.init_length = min(init_length, width//2)
        self.food_reward = food_reward
        self.dist_reward = (
            width+height)//4 if dist_reward is None else dist_reward
        self.living_bonus = living_bonus
        self.death_penalty = death_penalty
        self.blocks_x = width
        self.blocks_y = height
        self.food_color = food_color
        self.head_color = head_color
        self.body_color = body_color
        self.background_color = background_color
        self.food = Food(self.blocks_x, self.blocks_y, food_color)
        Block.size = block_size

        self.map = None
        self.screen = None
        self.clock = None
        self.human_playing = False

    def init(self):
        self.episode += 1
        self.score = 0
        self.direction = 3
        self.current_step = 0
        self.head = Block(self.blocks_x//2, self.blocks_y//2, self.head_color)
        self.body = [self.head.copy(i, 0, self.body_color)
                     for i in range(-self.init_length, 0)]
        self.blocks = [self.food.block, self.head, *self.body]
        self.food.new_food(self.blocks)

    def close(self):
        pygame.quit()
        pygame.display.quit()
        self.screen = None
        self.clock = None

    def render(self):
        if self.screen is None:
            self.screen, self.clock = game_start(
                self.blocks_x*Block.size, self.blocks_y*Block.size)
        self.clock.tick(self.fps)
        update_screen(self.screen, self)
        handle_input()

    def step(self, direction):
        if direction is None:
            direction = self.direction
        self.current_step += 1
        truncated = True if self.current_step == self.max_step else False
        (x, y) = (self.head.x, self.head.y)
        step = Direction.step(direction)
        if (direction == 0 or direction == 1) and (self.direction == 0 or self.direction == 1):
            step = Direction.step(self.direction)
        elif (direction == 2 or direction == 3) and (self.direction == 2 or self.direction == 3):
            step = Direction.step(self.direction)
        else:
            self.direction = direction
        self.head.x += step[0]
        self.head.y += step[1]

        reward = self.living_bonus + self.calc_reward()
        dead = False

        if self.head == self.food.block:
            self.score += 1
            self.grow(x, y)
            self.food.new_food(self.blocks)
            reward = self.food_reward
        else:
            self.move(x, y)
            for block in self.body:
                if self.head == block:
                    dead = True
            if self.head.x >= self.blocks_x or self.head.x < 0 or self.head.y < 0 or self.head.y >= self.blocks_x:
                dead = True
        if dead:
            reward = self.death_penalty
        return self.observation(dead), reward, dead, truncated

    def observation(self, dead=False):
        dx = self.head.x - self.food.block.x
        dy = self.head.y - self.food.block.y
        dx, dy = normalize(dx, dy)
        d0, d1, d2, d3 = self.calc_distance(dead)
        return np.array([dx, dy, d0, d1, d2, d3], dtype=np.float32)

    def calc_distance(self, dead):
        if dead:
            return 0, 0, 0, 0
        self.map = np.zeros((self.blocks_x, self.blocks_y), dtype=int)
        for block in self.blocks:
            self.map[block.x][block.y] = -1
        self.map[self.food.block.x][self.food.block.y] = 0
        d0, d1, d2, d3 = 0, 0, 0, 0,
        x, y = self.head.x, self.head.y - 1
        while y >= 0 and self.map[x][y] == 0:
            d0 += 1
            y -= 1
        x, y = self.head.x, self.head.y + 1
        while y < self.blocks_y and self.map[x][y] == 0:
            d1 += 1
            y += 1
        x, y = self.head.x - 1, self.head.y
        while x >= 0 and self.map[x][y] == 0:
            d2 += 1
            x -= 1
        x, y = self.head.x + 1, self.head.y
        while x < self.blocks_x and self.map[x][y] == 0:
            d3 += 1
            x += 1
        self.map[self.food.block.x][self.food.block.y] = 1
        return d0/self.blocks_y, d1/self.blocks_y, d2/self.blocks_x, d3/self.blocks_x

    def calc_reward(self):
        if self.dist_reward == 0.0:
            return 0
        x = self.head.x - self.food.block.x
        y = self.head.y - self.food.block.y
        d = math.sqrt(x*x + y*y)
        return (self.dist_reward-d)/self.dist_reward

    def grow(self, x, y):
        body = Block(x, y, Color.blue)
        self.blocks.append(body)
        self.body.append(body)

    def move(self, x, y):
        tail = self.body.pop(0)
        tail.move_to(x, y)
        self.body.append(tail)

    def info(self):
        return {
            'head': (self.head.x, self.head.y),
            'food': (self.food.block.x, self.food.block.y),
            'map': self.map.T
        }

    def play(self, fps=10, acceleration=True, step=1, frep=10):
        self.max_step = 99999
        self.fps = fps
        self.food_reward = 1
        self.living_bonus = 0
        self.dist_reward = 0
        self.death_penalty = 0
        self.human_playing = True
        self.init()
        screen, clock = game_start(
            self.blocks_x*Block.size, self.blocks_y*Block.size)
        total_r = 0
        while pygame.get_init():
            clock.tick(self.fps)
            _, r, d, _ = self.step(handle_input())
            total_r += r
            if acceleration and total_r == frep:
                self.fps += step
                total_r = 0
            if d:
                self.init()
                total_r = 0
                self.fps = fps
            update_screen(screen, self, True)



import pygame
import math

from random import randint


class Food:
    def __init__(self, blocks_x, blocks_y, color) -> None:
        self.blocks_x = blocks_x
        self.blocks_y = blocks_y
        self.block = Block(0, 0, color)

    def new_food(self, blocks):
        (x, y) = (randint(0, self.blocks_x-1), randint(0, self.blocks_y-1))
        done = False

        while not done:
            done = True
            for block in blocks:
                if (x, y) == (block.x, block.y):
                    (x, y) = (randint(0, self.blocks_x-1),
                              randint(0, self.blocks_y-1))
                    done = False
                    break
        self.block.move_to(x, y)


class Block:
    size = 20

    def __init__(self, x, y, color) -> None:
        self.x = x
        self.y = y
        self.color = color

    def copy(self, dx, dy, color=None):
        if color is None:
            color = self.color
        return Block(self.x+dx, self.y+dy, color)

    def __eq__(self, other) -> bool:
        return self.x == other.x and self.y == other.y

    def move_to(self, x, y):
        self.x = x
        self.y = y

    @property
    def rect(self):
        return (self.x*Block.size, self.y*Block.size, Block.size, Block.size)


class Color:
    red = (255, 90, 90)
    orange = (255, 169, 89)
    blue = (89, 172, 255)
    green_blue = (89, 255, 172)
    purple = (197, 90, 255)


class Direction:
    up = (0, -1)
    down = (0, 1)
    left = (-1, 0)
    right = (1, 0)

    @staticmethod
    def step(dir):
        match dir:
            case 0:
                return Direction.up
            case 1:
                return Direction.down
            case 2:
                return Direction.left
            case 3:
                return Direction.right


def normalize(x, y):
    c = math.sqrt(x*x + y*y)
    return x/c, y/c


def handle_input():
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            pygame.display.quit()
            pygame.quit()
            return None
    key_list = pygame.key.get_pressed()
    if key_list[pygame.K_ESCAPE]:
        pygame.display.quit()
        pygame.quit()
        return None
    key_list = pygame.key.get_pressed()
    u, d, l, r = key_list[pygame.K_UP] or key_list[pygame.K_w], \
        key_list[pygame.K_DOWN] or key_list[pygame.K_s], \
        key_list[pygame.K_LEFT] or key_list[pygame.K_a], \
        key_list[pygame.K_RIGHT] or key_list[pygame.K_d]
    if u:
        return 0
    if d:
        return 1
    if l:
        return 2
    if r:
        return 3
    return None


def update_screen(screen, snake, human_playing=False):
    if not pygame.display.get_init():
        return
    width = screen.get_width()
    height = screen.get_height() - 40
    font = pygame.font.SysFont('microsoft Yahei', 30, True)
    score = font.render('Scores: '+str(snake.score), False, Color.purple)
    episode = font.render('Episodes: '+str(snake.episode), False, Color.purple)
    step_remain = font.render(
        'Steps Remain: '+str(snake.max_step-snake.current_step), False, Color.purple)
    screen.fill(Color.orange)
    screen.blit(score, (20, height))
    if human_playing:
        fps = font.render('Speed: '+str(snake.fps), False, Color.purple)
        screen.blit(fps, (200, height))
    else:
        screen.blit(episode, (200, height))
        screen.blit(step_remain, (500, height))

    for block in snake.blocks:
        pygame.draw.rect(screen, block.color, pygame.Rect(block.rect))
    pygame.draw.line(screen, Color.purple, (0, height),
                     (width, height), 2)
    pygame.display.flip()


def game_start(width, height, score_board=40):
    pygame.init()
    pygame.display.set_caption('Snake')
    clock = pygame.time.Clock()
    screen = pygame.display.set_mode((width, height + score_board))
    return screen, clock


def calc_advantage(gamma, lmbda, td_delta):
    td_delta = td_delta.detach().numpy()
    advantage_list = []
    advantage = 0.0
    for delta in td_delta[::-1]:
        advantage = gamma * lmbda * advantage + delta
        advantage_list.append(advantage)
    advantage_list.reverse()
    return advantage_list



# https://github.com/NaLooo/Gym_Snake_Game

from setuptools import setup, find_packages
import io
from os import path

# --- get version ---
version = "unknown"
with open("gym_snake_game/version.py") as f:
    line = f.read().strip()
    version = line.replace("VERSION = ", "").replace("'", '')
# --- /get version ---

here = path.abspath(path.dirname(__file__))
with io.open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='gym_snake_game',
    version=version,
    description='Snake game for OpenAI Gym',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/NaLooo/Gym_Snake_Game',
    author='Ming Yu',
    author_email='ming.yu@alumni.stonybrook.edu',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Development Status :: 3 - Alpha',
        # 'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',

        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Games/Entertainment :: Arcade',

        'Programming Language :: Python :: 3.10',
    ],
    platforms=['any'],
    keywords='ai, rl, snake',
    packages=find_packages(),
    python_requires='>=3.0',
    install_requires=['pygame>=2.1.0', 'numpy>=1.15', 'gym>=0.26.2'],
)
